from django.db import models
from django.core.mail import get_connection, EmailMultiAlternatives, \
    EmailMessage
from jsonfield import JSONField
from django.template import Template, Context
import html2text
import re
from premailer import transform


class DeferredMail(models.Model):
    template = models.ForeignKey('EmailTemplate', null=True, blank=True, on_delete=models.CASCADE)
    local_template = models.CharField(max_length=255, null=True, blank=True)
    context = JSONField(blank=True, null=True)
    email = models.EmailField()
    title = models.CharField(max_length=255)
    key = models.CharField(max_length=100, db_index=True,
                           default='##default_key##')
    schedule_time = models.DateTimeField()

    def __str__(self):
        return "%s, %s" % (self.email, self.title)

    def __unicode__(self):
        return self.__str__()

    @classmethod
    def remove_by(cls, key):
        cls.objects.filter(key=key).delete()


class EmailTemplate(models.Model):
    name = models.SlugField()
    html_content = models.TextField()

    should_inline = models.BooleanField(default=False)

    @property
    def text_content(self):
        return html2text.html2text(self.html_content)

    def get_html_content(self):
        if self.should_inline:
            return transform(self.html_content)
        return self.html_content

    def __str__(self):
        return self.name

    def __unicode__(self):
        return self.__str__()

    def render(self, context):
        html_template = Template(self.get_html_content())
        text_template = Template(self.text_content)
        return html_template.render(Context(context)), text_template.render(
            Context(context))


class Provider(models.Model):
    enabled = models.BooleanField(default=True)
    smtp_host = models.CharField(max_length=255)
    smtp_port = models.IntegerField(default=587)
    smtp_username = models.CharField(max_length=255)
    smtp_password = models.CharField(max_length=255)
    use_tls = models.BooleanField(default=False)

    from_address = models.CharField(max_length=255)

    blacklist = JSONField(blank=True)

    quota = models.IntegerField(default=15000)
    usage = models.IntegerField(default=0)
    quota_type_is_daily = models.BooleanField(default=False)

    preference = models.IntegerField(default=0)

    def reset_usage(self):
        self.usage = 0
        self.save()

    def add_usage(self):
        self.usage += 1
        self.save()

    def within_quota(self):
        if not self.quota:
            return True
        return self.usage < self.quota

    def __str__(self):
        return self.smtp_host

    def __unicode__(self):
        return self.__str__()

    def get_connection(self):
        return get_connection(host=self.smtp_host,
                              port=self.smtp_port,
                              username=self.smtp_username,
                              password=self.smtp_password,
                              use_tls=self.use_tls)

    def send(self, address, title, content, html_message=None,
             attachments=None):
        msg = EmailMessage(title, content, self.from_address, [address],
                           connection=self.get_connection())

        if html_message or attachments:
            msg = EmailMultiAlternatives(title, content, self.from_address,
                                         [address],
                                         connection=self.get_connection()
                                         )
            if html_message:
                msg.attach_alternative(html_message, "text/html")
            if attachments:
                for attachment in attachments:
                    msg.attach_file(attachment)

        msg.send()
        msg.connection.close()
        self.add_usage()

    def can_send(self, address):
        if not self.enabled:
            return False
        if not self.within_quota():
            return False
        suffix = re.search("@[\w.]+", address).group()
        if self.blacklist and suffix in self.blacklist:
            return False
        return True



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-17 01:06
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('asyncmailer', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DeferredMail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('context', jsonfield.fields.JSONField(blank=True, null=True)),
                ('email', models.EmailField(max_length=254)),
                ('title', models.CharField(max_length=255)),
                ('key', models.CharField(db_index=True, default='##default_key##', max_length=100)),
                ('schedule_time', models.DateTimeField()),
            ],
        ),
        migrations.CreateModel(
            name='EmailTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField()),
                ('html_content', models.TextField()),
                ('text_content', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Provider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('enabled', models.BooleanField(default=True, verbose_name=b'If this provider is enabled')),
                ('smtp_host', models.CharField(max_length=255, verbose_name=b'Host of SMTP Server')),
                ('smtp_port', models.IntegerField(default=587, verbose_name=b'Port of SMTP Server')),
                ('smtp_username', models.CharField(max_length=255, verbose_name=b'User Name of SMTP Server')),
                ('smtp_password', models.CharField(max_length=255, verbose_name=b'Password of SMTP Server')),
                ('use_tls', models.BooleanField(default=False, verbose_name=b'If the server uses TLS')),
                ('from_address', models.CharField(max_length=255, verbose_name=b'From address for this SMTP Server')),
                ('blacklist', jsonfield.fields.JSONField(blank=True, verbose_name=b'Never send emails to these domains with this provider. e.g., @msn.com')),
                ('quota', models.IntegerField(default=15000, verbose_name=b'Quota of the SMTP Server')),
                ('usage', models.IntegerField(default=0, verbose_name=b'Usage of the SMTP Server')),
                ('quota_type_is_daily', models.BooleanField(default=False, verbose_name=b'Is the quota of the server daily?')),
                ('preference', models.IntegerField(default=0, verbose_name=b'Preference of the Server, Higher the Better')),
            ],
        ),
        migrations.AddField(
            model_name='deferredmail',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='asyncmailer.EmailTemplate'),
        ),
    ]



from asyncmailer.tasks import async_select_and_send
from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http import JsonResponse
from django.shortcuts import render
from django.template.loader import render_to_string
from premailer import transform
import html2text
import json
import os
import re


def name_to_file(template_name, variation=None):
    if variation:
        return "%s-templates/%s.json" % (template_name, variation)
    return "%s-templates/%s.html" % (template_name, template_name)


def find_templates():
    template_path = []

    if hasattr(settings, "TEMPLATE_DIRS"):
        template_path += settings.TEMPLATE_DIRS

    if hasattr(settings, "TEMPLATES") and settings.TEMPLATES:
        template_path += settings.TEMPLATES[0]['DIRS']
    template_path += [os.path.dirname(__file__) + '/templates/']
    template_path = list(set(template_path))
    templates = {}
    for path in template_path:
        for dirname in os.listdir(path):
            if not os.path.isdir(path + dirname) or dirname == 'asyncmailer':
                continue
            else:
                template_name = ""
                new_template = {"dir": dirname, "html": "", "variations": []}
                for filename in os.listdir(os.path.join(path, dirname)):
                    if re.findall('\.html$', filename):
                        template_name = filename[:-5]  # Drop .html
                    elif re.findall('\.json$', filename):
                        new_template['variations'].append(filename[:-5])
                        # Drop .json
                    else:
                        pass
                templates[template_name] = new_template
    return templates


def get_request_content(request):
    template = request.POST.get('template', '')
    variation = request.POST.get('variation', '')
    locale = request.POST.get('locale', '')
    inline = request.POST.get('inline', '') == 'On'
    formats = request.POST.get('formats', '')
    payload = request.POST.get('payload', '{}')
    payload = json.loads(payload)
    return template, variation, locale, inline, formats, payload


@staff_member_required
def index(request):
    return render(request, 'asyncmailer/index.html', {
        'templates': find_templates().keys()
    })


@staff_member_required
def get_variations(request):
    template = request.POST.get('template', '')
    variations = find_templates()[template]["variations"]
    return JsonResponse({"variations": variations})


@staff_member_required
def get_metadata(request):
    template = request.POST.get('template', '')
    variation = request.POST.get('variation', '')
    response = {}
    if variation != 'base.json':
        base_json = render_to_string(
            name_to_file(template, variation="base")
        )
        response.update(json.loads(str(base_json)))
    variation_json = render_to_string(
        name_to_file(template, variation=variation)
    )
    response.update(json.loads(str(variation_json)))
    return JsonResponse(response)


@staff_member_required
def render_email(request):
    try:
        template, variation, locale, inline, formats, payload = \
            get_request_content(request)
    except:
        return JsonResponse({
            'success': False,
            'error': "Invalid JSON in payload."
        })

    try:
        rich_text = render_to_string(
            name_to_file(template), payload)
        plain_text = html2text.html2text(rich_text)
        if inline:
            rich_text = transform(rich_text)
        if formats == 'text':
            rich_text = None

    except:
        return JsonResponse({
            'success': False,
            'error': "Template render error."
        })
    return JsonResponse({
        'success': True,
        'content': rich_text or plain_text,
        'subject': payload["subject"],
        'plainText': formats == 'text'
    })


@staff_member_required
def send_test_mail(request):
    try:
        template, variation, locale, inline, formats, payload = \
            get_request_content(request)
    except ValueError:
        return JsonResponse({
            'success': False,
            'error': "Invalid JSON in payload."
        })
    try:
        emails = request.POST.get('email', None)
        if not emails:
            raise ValueError
        emails = [e.strip() for e in emails.split(',')]
    except:
        return JsonResponse({
            'success': False,
            'error': "Invalid email address."
        })

    try:
        rich_text = render_to_string(name_to_file(template), payload)
        plain_text = html2text.html2text(rich_text)
        if inline:
            rich_text = transform(rich_text)
        if formats == 'text':
            rich_text = None
    except:
        return JsonResponse({
            'success': False,
            'error': "Template render error."
        })

    try:
        for email in emails:
            async_select_and_send(email, payload["subject"], plain_text,
                                  rich_text=rich_text)
    except:
        return JsonResponse({
            'success': False,
            'error': "Cannot send email. Check Provider settings."
        })
    return JsonResponse({'success': True})



"""URLs to run the tests."""
from compat import include, url
from django.contrib import admin
import asyncmailer.urls


admin.autodiscover()

urlpatterns = [
    url(r'^asyncmailer/', include(asyncmailer.urls)),
    url(r'^admin/', include(admin.site.urls)),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-03-13 21:03
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('asyncmailer', '0003_auto_20160817_0107'),
    ]

    operations = [
        migrations.AddField(
            model_name='deferredmail',
            name='local_template',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='deferredmail',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='asyncmailer.EmailTemplate'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-17 01:05
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
    ]



"""Templatetags for the asyncmailer app."""
from django import template

register = template.Library()

# @register.filter
# def lower(value):
#     """
#     Converts a string into all lowercase
#
#     """
#     return value.lower()



from celery import shared_task
from django.template.loader import render_to_string
from celery.schedules import crontab
from celery.task import periodic_task
from django.utils import timezone
from asyncmailer.models import Provider, EmailTemplate, DeferredMail
import html2text
import random


@shared_task(default_retry_delay=5, max_retries=3)
def async_select_and_send(email, title, plain_text, rich_text=None,
                          attachments=None, **kwargs):
    try:
        providers = Provider.objects.all()
        good_providers = [x for x in providers if x.can_send(email)]
        top_preference = good_providers[0].preference
        top_providers = [provider for provider in good_providers if
                         provider.preference == top_preference]
        selected_provider = random.choice(top_providers)
        selected_provider.send(email, title, plain_text, rich_text,
                               attachments=attachments)
    except Exception as exc:
        raise async_select_and_send.retry(exc=exc)


def async_mail(email, title, context_dict=None, attachments=None,
               template='email-templates/email.html'):
    if len(email) == 1:
        rich_text = render_to_string(template, context_dict)
        plain_text = html2text.html2text(rich_text)
        async_select_and_send.delay(email[0], title, plain_text,
                                    rich_text, attachments=attachments)
    else:
        for address in email:
            async_mail(address, title, context_dict=context_dict[address],
                       attachments=attachments, template=template)


def add_deferred_mail(email, title, template_name, key, delta,
                      context_dict=None, local_template=None):
    now = timezone.now()
    schedule_time = now + delta
    template = EmailTemplate.objects.get(name=template_name) \
        if template_name else None
    m = DeferredMail(
        template=template,
        local_template=local_template,
        context=context_dict,
        email=email,
        title=title,
        key=key,
        schedule_time=schedule_time
    )
    m.save()


def remove_deferred_mail(key):
    DeferredMail.remove_by(key)


@periodic_task(run_every=crontab(minute=10))
def send_deferred_mails():
    for mail in DeferredMail.objects.filter(schedule_time__lt=timezone.now()):
        if mail.template:
            html_content, text_content = mail.template.render(mail.context)
        else:
            html_content = render_to_string(
                mail.local_template,
                mail.context,
            )
            text_content = html2text.html2text(html_content)

        async_select_and_send.delay(mail.email,
                                    mail.title,
                                    text_content,
                                    html_content)
        mail.delete()


@periodic_task(run_every=crontab(hour=0, minute=0))
def clear_daily_usages():
    providers = Provider.objects.filter(quota_type_is_daily=True)
    for p in providers:
        p.reset_usage()


@periodic_task(run_every=crontab(day_of_month=1, hour=0, minute=0))
def clear_monthly_usages():
    providers = Provider.objects.filter(quota_type_is_daily=False)
    for p in providers:
        p.reset_usage()



"""URLs for the asyncmailer app."""
from compat import url
from . import views

urlpatterns = [
    url(r'^$', views.index),
    url(r'^index$', views.index),
    url(r'^get_variations$', views.get_variations),
    url(r'^get_metadata$', views.get_metadata),
    url(r'^render_email$', views.render_email),
    url(r'^send_test_mail$', views.send_test_mail),
    # append your urls here
]



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-17 01:07
from __future__ import unicode_literals

from django.db import migrations, models
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('asyncmailer', '0002_auto_20160817_0106'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='emailtemplate',
            name='text_content',
        ),
        migrations.AddField(
            model_name='emailtemplate',
            name='should_inline',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='provider',
            name='blacklist',
            field=jsonfield.fields.JSONField(blank=True),
        ),
        migrations.AlterField(
            model_name='provider',
            name='enabled',
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='provider',
            name='from_address',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='provider',
            name='preference',
            field=models.IntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='provider',
            name='quota',
            field=models.IntegerField(default=15000),
        ),
        migrations.AlterField(
            model_name='provider',
            name='quota_type_is_daily',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='provider',
            name='smtp_host',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='provider',
            name='smtp_password',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='provider',
            name='smtp_port',
            field=models.IntegerField(default=587),
        ),
        migrations.AlterField(
            model_name='provider',
            name='smtp_username',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='provider',
            name='usage',
            field=models.IntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='provider',
            name='use_tls',
            field=models.BooleanField(default=False),
        ),
    ]



"""Tests for the models of the asyncmailer app."""
# from django.test import TestCase

# from mixer.backend.django import mixer
from django.test import TestCase
from asyncmailer.models import Provider, EmailTemplate
from asyncmailer.tasks import clear_daily_usages, clear_monthly_usages


# Create your tests here.
class ProviderTest(TestCase):
    """SMTP Provider Test"""

    def setUp(self):
        self.test_provider = Provider(
            enabled=True,
            smtp_host="localhost",
            smtp_port=1025,
            smtp_username="",
            smtp_password="",
            use_tls=False,
            from_address="test@example.com",
            blacklist=[]
        )

    def test_can_send_normal(self):
        """
        Test if the provider can send a normal mail
        """
        self.assertEqual(self.test_provider.can_send("test@example.com"), True)

    def test_disabled(self):
        """
        Test if the provider works when disabled
        """
        self.test_provider.enabled = False
        self.assertEqual(self.test_provider.can_send("test@example.com"),
                         False)
        self.test_provider.enabled = True

    def test_blacklist(self):
        """
        Test if the provider can send a mail to an address in the blacklist.
        """
        self.test_provider.blacklist = ["@qq.com"]
        self.assertEqual(self.test_provider.can_send("test@qq.com"), False)

    def test_over_quota(self):
        """
        Test if the provider works when out of quota
        """
        pre_usage = self.test_provider.usage
        self.test_provider.usage = self.test_provider.quota
        self.assertEqual(self.test_provider.can_send("test@example.com"),
                         False)
        self.test_provider.usage = pre_usage

    def test_reset_usage(self):
        """
        Test if the provider resets usage
        """
        self.test_provider.usage = 10
        self.test_provider.reset_usage()
        self.assertEqual(self.test_provider.usage, 0)

    def test_usage_count(self):
        """
        Test if the provider works with usage counting
        """
        pre_usage = self.test_provider.usage
        self.test_provider.send("test@example.com", "Test", "Test",
                                html_message="<html></html>")
        post_usage = self.test_provider.usage
        self.assertEqual(post_usage - pre_usage, 1)

    def test_clear_daily_usages(self):
        """
        Test if clear_daily_usages works
        """
        self.test_provider.quota_type_is_daily = True
        self.test_provider.usage = 100
        self.test_provider.save()
        clear_daily_usages()
        self.test_provider = Provider.objects.get(pk=self.test_provider.pk)
        self.assertEqual(self.test_provider.usage, 0)

    def test_clear_monthly_usages(self):
        """
        Test if clear_monthly_usages works
        """
        self.test_provider.quota_type_is_daily = False
        self.test_provider.usage = 100
        self.test_provider.save()
        clear_monthly_usages()
        self.test_provider = Provider.objects.get(pk=self.test_provider.pk)
        self.assertEqual(self.test_provider.usage, 0)

    def test_template_render(self):
        """
        Test if the template renders correctly
        """
        original = """\
<style>
h1 {
    font-family: serif;
}
</style>
<h1>Title {{ name }}</h1>
<p>Content</p>"""

        template = EmailTemplate(
            name="Test Template",
            html_content=original,
            should_inline=True
        )
        context = {
            "name": "Andy"
        }
        should_rich = """\
<html>
<head></head>
<body>
<h1 style="font-family:serif">Title Andy</h1>
<p>Content</p>
</body>
</html>
"""
        should_plain = "# Title Andy\n\nContent\n\n"
        rich, plain = template.render(context=context)
        self.assertEqual(rich, should_rich)
        self.assertEqual(plain, should_plain)
